// Copyright 2013 Krzysztof Czainski
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

/*
 * @file boost/move/core2.hpp
 *
 * @since 2013-01-04
 * @author  Krzysztof Czainski <1czajnik@gmail.com>
 */

#ifndef HPP_BOOSTMOVE_CORE2_
#define HPP_BOOSTMOVE_CORE2_

#include <boost/move/core.hpp>

#ifdef DOXYGEN

/// @addtogroup boost_move
//@{

/// Use this macro to make your class movable.
///
/// When using this macro in a class:
/// - provide a non-member `noexcept` `swap()` for your class;
/// - do **not** provide a copy-assignment operator, because it's provided for you, implemented
///   in terms of pass-by-value and swap;
/// - specify a variant of the copy constructor in one of the following scenarios:
///   - using `BOOST_DELETED_COPY_CONSTRUCTOR(TYPE)` if you want a movable-only class;
///   - using `BOOST_EXPLICIT_COPY_CONSTRUCTOR(TYPE,x)` if you want an explicitly-copyable class;
///     in C++98 your class won't have an `explicit` copy constructor and you can copy instances
///     using `boost::copy()`, but classes containing or inheriting from your class won't have
///     auto-generated copy constructors;
///   - you might want to provide a copy constructor, that is `explicit` in C++11, and implicit in
///     C++98 (so that classes containing or inheriting from your class will have auto-generated
///     copy constructors); it is best to show that directly in your code:
///     @code
///     #ifndef BOOST_NO_RVALUE_REFERENCES
///       explicit
///     #endif
///       TYPE( TYPE const& b ) : ... { ... }
///     @endcode
///     this can lead to some undetected implicit copies of your class in C++98, which won't
///     compile in C++11 and can be fixed trivially using `boost::copy()`; **attention:**
///     do not provide an `explicit` copy constructor in C++98, because it is dangerous and
///     together with move emulation makes your class behave like the deprecated `std::auto_ptr`;
///   - provide an ordinary copy constructor to make your class implicitly copyable and movable.
///
/// If your move constructor is fast and `noexcept`, it is recommended to make the copy
/// constructor `explicit` (see `boost::copy()`), otherwise `swap` probably isn't fast or
/// `noexcept` either, in which case you shouldn avoid the copy-and-swap idiom, used in the
/// implementation of this macro.
///
/// @attention It is unspecified, which access specifier is left after this macro.
///
/// @snippet TODO
#define BOOST_MOVABLE2(TYPE)

/// This macro disables the copy constructor in a way compatible with `BOOST_MOVABLE2(TYPE)` for
/// your compiler.
///
/// @attention It is unspecified, which access specifier is left after this macro.
///
/// @snippet TODO
#define BOOST_DELETED_COPY_CONSTRUCTOR(TYPE)

/// This macro declares an `explicit` copy constructor in a way compatible with
/// `BOOST_MOVABLE2(TYPE)` for your compiler; you need to provide its' implementation.
///
/// In C++98 your class won't have an `explicit` copy constructor and you can copy instances
/// using `boost::copy()`, but classes containing or inheriting from your class won't have
/// auto-generated copy constructors.
///
/// @attention It is unspecified, which access specifier is left after this macro.
///
/// @snippet TODO
#define BOOST_EXPLICIT_COPY_CONSTRUCTOR(TYPE,x)

/// This macro should be used whenever you return a local variable.
///
/// In C++11 it expands to `(x)`, while in C++11 it expands to `::boost::move(x)`. It's needed,
/// because:
/// - in C++11 you shouldn't wrap local returned variables with `move()` (moving will happen
///   automatically, and wrapping with `move()` may disable some move-elision opportunities;
/// - in C++98 the `move()` is necessary.
#define BOOST_MOVE_LOCAL_RETVAL(x)

//@}

#else // DOXYGEN

#ifdef BOOST_NO_CXX11_DELETED_FUNCTIONS

#define BOOST_IMPL_DELETED_COPY_CONSTRUCTOR(TYPE) \
private:\
    TYPE(TYPE &);

#else

#define BOOST_IMPL_DELETED_COPY_CONSTRUCTOR(TYPE) \
public:\
    TYPE(TYPE const &) = delete;

#endif // BOOST_NO_CXX11_DELETED_FUNCTIONS

#if defined(BOOST_NO_RVALUE_REFERENCES)

#define BOOST_DELETED_COPY_CONSTRUCTOR(TYPE) \
    BOOST_IMPL_DELETED_COPY_CONSTRUCTOR(TYPE)

#define BOOST_MOVABLE2(TYPE) \
public: \
    operator ::boost::rv<TYPE>&() \
    { return *static_cast< ::boost::rv<TYPE>* >(this); } \
    operator const ::boost::rv<TYPE>&() const \
    { return *static_cast<const ::boost::rv<TYPE>* >(this); } \
    TYPE& operator=( TYPE b ) BOOST_NOEXCEPT { swap(*this,b); return *this; }

namespace boost { namespace move_detail {

struct explicit_copy_constructor_tag {};

}}  // boost::move_detail

#define BOOST_EXPLICIT_COPY_CONSTRUCTOR(TYPE,x) \
    BOOST_DELETED_COPY_CONSTRUCTOR(TYPE) \
public: \
    friend TYPE copy_adl( TYPE const& x ) \
    { return TYPE( x, boost::move_detail::explicit_copy_constructor_tag() ); } \
private: \
    explicit TYPE( TYPE const& x, boost::move_detail::explicit_copy_constructor_tag )

#define BOOST_MOVE_LOCAL_RETVAL(x) ::boost::move(x)

#else // defined(BOOST_NO_RVALUE_REFERENCES)

#define BOOST_MOVABLE2(TYPE) \
public: \
    TYPE& operator=( TYPE b ) BOOST_NOEXCEPT { swap(*this,b); return *this; }

#define BOOST_DELETED_COPY_CONSTRUCTOR(TYPE) \
public: \
    typedef int boost_move_emulation_t; \
    BOOST_IMPL_DELETED_COPY_CONSTRUCTOR(TYPE)

#define BOOST_EXPLICIT_COPY_CONSTRUCTOR(TYPE,x) \
public: \
    explicit TYPE( TYPE const& x )

#define BOOST_MOVE_LOCAL_RETVAL(x) (x)

#endif // defined(BOOST_NO_RVALUE_REFERENCES)

#endif // DOXYGEN

#endif // HPP_BOOSTMOVE_CORE2_
