// Copyright 2014 Krzysztof Czainski
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

/*
 * @file  CloneablePtr.cpp
 *
 * @since  14 lut 2014
 * @author  Krzysztof Czainski <1czajnik@gmail.com>
 */

#include <boost/move/copy.hpp>

#include <boost/ptr_container/clone_allocator.hpp>

#include <memory>
#include <type_traits>
#include <utility>

/// [delete]
template < class T, class CloneAllocator = boost::heap_clone_allocator >
struct CloneableDelete : CloneAllocator
{
    typedef CloneAllocator clone_allocator;

    CloneableDelete() {}

    /// Conversion from other clone_allocator_delete, whose pointers are convertible.
    template < class U >
    CloneableDelete( CloneableDelete<U,CloneAllocator> const&,
            typename std::enable_if< std::is_convertible<U*,T*>::value >::type* = 0 )
    {}

    /// Invokes `deallocate_clone(ptr)`.
    void operator()( T* ptr ) const
    { this->deallocate_clone(ptr); }
};
/// [delete]

/// [CloneablePtr]
template < class T, class CloneAllocator = boost::heap_clone_allocator >
using CloneablePtr = std::unique_ptr< T, CloneableDelete<T,CloneAllocator> >;

template < class T, class... Args >
inline
CloneablePtr<T> make_Cloneable( Args&&... args )
{ return CloneablePtr<T>{ new T( std::forward<Args>(args)... ) }; }
/// [CloneablePtr]

/// [copy]
template < class T, class CloneAllocator >
CloneablePtr<T,CloneAllocator> copy_adl( CloneablePtr<T,CloneAllocator> const& x )
{
    typedef CloneablePtr<T,CloneAllocator> Ptr;
    return
        x ?
            Ptr( x.get_deleter().allocate_clone(*x), x.get_deleter() )
        :
            Ptr( 0, x.get_deleter() );
        ;
}
/// [copy]

/// [main]
struct A
{
    A() { std::cout << "A()" << std::endl; }
    A( A const& ) { std::cout << "A(c&)" << std::endl; }
    virtual ~A() { std::cout << "~A()" << std::endl; }
    virtual A* clone() const { return new A(*this); }
};

// used via ADL by boost::heap_clone_allocator
A* new_clone( A const& a )
{ return a.clone(); }

struct B : A
{
    B() { std::cout << "B()" << std::endl; }
    B( B const& b ) : A(b) { std::cout << "B(c&)" << std::endl; }
    ~B() { std::cout << "~B()" << std::endl; }
    B* clone() const { return new B(*this); }
};

int main()
{
    CloneablePtr<A> a0( new A );    // A()
    std::cout << "-- a0 contains an A" << std::endl;
                                    // -- a0 contains an A

    a0 = make_Cloneable<A>();       // A()
                                    // ~A();
    std::cout << "-- a0 contains an other A" << std::endl;
                                    // -- a0 contains an other A

    auto a1 = boost::copy(a0);      // A(c&)
    std::cout << "-- a1 contains a copy of a0's A" << std::endl;
                                    // -- a1 contains a copy of a0's A

    a0 = make_Cloneable<B>();       // A()
                                    // B()
                                    // ~A()
    std::cout << "-- a0 contains a B" << std::endl;
                                    // -- a0 contains a B

    auto a2 = boost::copy(a0);      // A(c&)
                                    // B(c&)
    std::cout << "-- a2 contains a copy of a0's B" << std::endl;
                                    // -- a2 contains a copy of a0's B

                                    // ~B()
                                    // ~A()
                                    // ~A()
                                    // ~B()
                                    // ~A()
}
/// [main]
